# -*- coding:utf-8 _*-  
"""
@author: Maureen Hsu
@file: layers.py 
@time: 2020/06/16
"""

# python packages

# 3rd-party packages
from loguru import logger
from tensorflow.keras.layers import (
    Dense,
    Conv2D, SeparableConv2D, DepthwiseConv2D,
    Add, Concatenate, Reshape, multiply, Flatten,
    GlobalAveragePooling2D, Activation)
from tensorflow.keras.layers import LeakyReLU
from tensorflow.keras.layers import BatchNormalization
from tensorflow.keras.regularizers import l2

# self-defined packages
from cv.module.utils import compose

kernel_reg = l2(5e-5)  # regularizer weight

"""
Elementary Layer Definition
"""


def Conv2DBasic(*args, **kwargs):
    default_kwargs = {'strides': (1, 1),
                      'kernel_regularizer': kernel_reg,
                      'padding': 'same',
                      'kernel_initializer': 'he_uniform'}
    default_kwargs.update(kwargs)
    return Conv2D(*args, **default_kwargs)


def SeparableConvBasic(*args, **kwargs):
    default_kwargs = {'strides': (1, 1),
                      'pointwise_regularizer': kernel_reg,
                      'padding': 'same',
                      'depthwise_initializer': 'he_uniform'}
    default_kwargs.update(kwargs)
    return SeparableConv2D(*args, **default_kwargs)


def DepthwiseConvBasic(*args, **kwargs):
    default_kwargs = {'strides': (1, 1),
                      'depthwise_regularizer': kernel_reg,
                      'padding': 'same',
                      'depthwise_initializer': 'he_uniform'}
    default_kwargs.update(kwargs)
    return DepthwiseConv2D(*args, **default_kwargs)


"""
Combined Layer Definition
"""


def ConvolutionBN(*args, **kwargs):
    no_bias_kwargs = {'use_bias': False}
    no_bias_kwargs.update(kwargs)
    return compose(
        Conv2DBasic(*args, **kwargs),
        BatchNormalization(),
        LeakyReLU(alpha=0.1))


def SeparableConvolutionBN(*args, **kwargs):
    no_bias_kwargs = {'use_bias': False}
    no_bias_kwargs.update(kwargs)
    return compose(
        SeparableConvBasic(*args, **kwargs),
        BatchNormalization(),
        LeakyReLU(alpha=0.1))


def DepthwiseConvolutionBNLayer(*args, **kwargs):
    no_bias_kwargs = {'use_bias': False}
    no_bias_kwargs.update(kwargs)
    return compose(
        DepthwiseConvBasic(*args, **kwargs),
        BatchNormalization(),
        LeakyReLU(alpha=0.1))


"""
SOTA module Definition
"""


def mobilenet_v2_reduce_body(input, out_channels, expansion_factor=2):
    in_channels = input.get_shape().as_list()[3]

    x = ConvolutionBN(in_channels * expansion_factor, kernel_size=(1, 1))(input)
    x = DepthwiseConvolutionBNLayer(kernel_size=(3, 3), strides=(2, 2))(x)
    x = Conv2DBasic(out_channels, kernel_size=(1, 1))(x)
    return x


def mobilenet_v2_body(input, expansion_factor=2):
    in_channels = input.get_shape().as_list()[3]

    x = ConvolutionBN(in_channels * expansion_factor, kernel_size=(1, 1))(input)
    x = DepthwiseConvolutionBNLayer(kernel_size=(3, 3))(x)
    x = Conv2DBasic(in_channels, kernel_size=(1, 1))(x)
    x = Add()([x, input])
    return x


def RecognizeLayer(x):
    x_number = Dense(128)(x)
    x_number = BatchNormalization()(x_number)
    x_number = LeakyReLU(alpha=0.1)(x_number)
    x_number = Dense(7)(x_number)
    return x_number


def ColorLayer(x):
    x = GlobalAveragePooling2D()(x)
    x_color = Dense(128)(x)
    x_color = BatchNormalization()(x_color)
    x_color = LeakyReLU(alpha=0.1)(x_color)
    x_color = Dense(2, activation='softmax', name="color")(x_color)
    return x_color


def AttentionRoiLayer(_input, x_color, ratio=16):
    conv_shape = _input.get_shape().as_list()
    h, w, filters = conv_shape[1:]

    x = ConvolutionBN(filters=filters // ratio, kernel_size=(1, 1))(_input)
    x = Flatten()(x)
    x = Concatenate(axis=-1)([x, x_color])
    x = Dense(h * w * 2)(x)
    x = BatchNormalization()(x)
    x = Activation('relu')(x)
    x = Dense(h * w)(x)
    x = BatchNormalization()(x)
    x = Activation('sigmoid')(x)
    x = Reshape((h, w, 1))(x)

    x = multiply([_input, x])
    return x


def AttentionLayer(_input, ratio=16):
    """
    SE Net: global pooling -> fc -> relu -> fc -> sigmoid ---|
                |                                            |
                --------------------------------------------->  scale
    :param input:
    :param ratio:
    :return:
    """
    conv_shape = _input.get_shape().as_list()
    filters = conv_shape[-1]

    x = GlobalAveragePooling2D()(_input)
    x = Reshape((1, 1, filters))(x)
    x = Dense(filters // ratio)(x)
    x = BatchNormalization()(x)
    x = Activation('relu')(x)
    x = Dense(filters)(x)
    x = Activation('sigmoid')(x)
    x = multiply([_input, x])
    return x



# -*- coding:utf-8 _*-  
"""
@author: Maureen Hsu
@file: data_gen.py 
@time: 2020/03/09
"""

# python packages

# 3rd-party packages
from loguru import logger
import numpy as np
import tensorflow as tf
import tensorflow.keras as keras


# self-defined packages


@logger.catch(reraise=True)
class VOCDataGenerator(keras.utils.Sequence):
    """Generates data for Keras"""

    def __init__(self, dataset, batch_size,
                 read_data_method, data_process_method=None, image_aug=None,
                 shuffle=True, loop=True):
        """Initialization"""
        self.dataset = dataset  # VOCDataSplitLabels
        self.read_data_method = read_data_method
        self.data_process_method = data_process_method
        self.image_aug = image_aug
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.loop = loop
        self.indexes = np.arange(len(self.dataset))
        self.on_epoch_end()

    def __len__(self):
        """Denotes the number of batches per epoch"""
        return len(self.dataset) // self.batch_size

    def __getitem__(self, index):
        """Generate one batch of data"""
        # Generate data
        X, y = self.__data_generation(index)

        return X, y

    def on_epoch_end(self):
        self.indexes = np.arange(len(self.dataset))
        """Updates indexes after each epoch"""
        if self.shuffle is True:
            np.random.shuffle(self.indexes)

    def __data_generation(self, index):
        """Generates data containing batch_size samples"""  # X : (n_samples, *dim, n_channels)
        i = self.batch_size * index
        batch_images, batch_labels = [], []
        while len(batch_images) < self.batch_size:
            batch_indexes = []
            for b in range(self.batch_size):
                batch_indexes.append(self.indexes[i])

                i = (i + 1) % self.dataset.nb_samples

            # read data
            images, labels = self.read_data_method(self.dataset.annotations, batch_indexes)

            # append to batch
            batch_images += images
            batch_labels += labels

        # data augmentation
        if self.image_aug is not None:
            batch_images, batch_labels = self.image_aug.aug(batch_images, batch_labels)

        # process data
        if self.data_process_method is not None:
            batch_images, batch_labels = self.data_process_method(batch_images, batch_labels)

        return batch_images, batch_labels

    def get_all(self):
        images, labels = self.read_data_method(self.dataset.annotations, self.indexes)
        return images, labels



# -*- coding:utf-8 _*-  
"""
@author: Maureen Hsu
@file: __init__.py.py 
@time: 2019/10/25
"""

# python packages

# 3rd-party packages
# from configobj import ConfigObj
# from loguru import logger

# self-defined packages
# from utils import os_path

# global variables
# log_path = "./log/"
# config_path = "./config/"

# set up log & config
# logger.add(log_path, rotation="5 MB", retention="1 week") # filter=lambda record: record["extra"].get("name") == ""
# logger = logger.bind(name="")
# config = ConfigObj(config_path)


# -*- coding:utf-8 _*-  
"""
@author: Maureen Hsu
@file: model.py 
@time: 2020/06/01
"""

# python packages

# 3rd-party packages
import imgaug as ia
import numpy as np
import pandas as pd
from tensorflow.keras.models import load_model, Model
from loguru import logger

# self-defined packages
from cv.yolov3.core.models import get_anchor_masks
from cv.yolov3.core.models import (
    YoloCustomizeModelMultiCategory
)
from cv.yolov3.core.utils import (
    decode_nms,
    convert2_class
)
from utils import os_path, image_processing
from utils.label.VOCLabel import (
    VOCAnnotation,
    VOCObject
)


@logger.catch(reraise=True)
class ImageAug:
    def __init__(self, aug_seq=None, resize_seq=None):
        self.aug_seq = aug_seq
        self.resize_seq = resize_seq

    def aug(self, images, labels=None):
        seqs = []
        if self.aug_seq:
            seqs.append(self.aug_seq)

        if self.resize_seq:
            seqs.append(self.resize_seq)

        for seq in seqs:
            seq = seq.to_deterministic()
            if not labels:
                images = seq.augment_images(images)
            else:
                images, labels = seq(images=images, bounding_boxes=labels)
        return images, labels


class CNNModel(Model):
    def __init__(self, model_name, input_size):
        super(CNNModel, self).__init__()
        input_size = tuple(map(int, input_size))
        self.input_size = input_size
        self.model_name = model_name

    def load(self, model_dir=None, model_path=None):
        if model_dir is None and model_path is None:
            raise Exception("No model path can be loaded")
        if model_path is None:
            model_path = os_path.join(model_dir, f"{self.model_name}.h5")
        if not os_path.exists(model_path):
            logger.warning("Pretrained model does not exist in {}.".format(model_path))
            return
        logger.debug('Load model {}.'.format(model_path))
        model = load_model(model_path, compile=False)
        self.__dict__.update(model.__dict__)

    def load_weights(self, weight_dir=None, weight_path=None, force=False, **kwargs):
        if not force:
            if weight_dir is None and weight_path is None:
                raise Exception("No model_weight path can be loaded")
            if weight_path is None:
                weight_path = os_path.join(weight_dir, f"{self.model_name}.h5")
            if not os_path.exists(weight_path):
                logger.warning("Pretrained weights does not exist in {}.".format(weight_path))
                return

        logger.debug('Load weights {}.'.format(weight_path))
        super(CNNModel, self).load_weights(weight_path, **kwargs)


class ClassificationModel(CNNModel):
    def __init__(self, name, input_size, list_class_names):
        super().__init__(name, input_size)
        self.list_class_names = list_class_names

    def get_pred_df(self, *args, **kwargs):
        raise NotImplementedError

    def predicted_to_klasses(self, pred):
        """
        :param pred: scores of each classes [np.array(n_sample), np.array(n_sample)... len(list_class_names)]
        :return: [[klass_1, klass_2], [klass_1, klass_2], ...n_sample]
        """
        for n, klass_name in enumerate(self.list_class_names):
            pred[n] = np.argmax(pred[n], axis=-1)
        klasses = []
        for i in range(len(pred[0])):
            klasses_of_one_sample = []
            for n, klass_name in enumerate(self.list_class_names):
                klasses_of_one_sample.append(klass_name[pred[n][i]])
            klasses.append(klasses_of_one_sample)
        return klasses

    def test(self, gen_test, output_dir=None):
        if self.model is None:
            raise Exception("Model is not defined.")

        images, labels = gen_test.get_all()

        logger.debug("Predicting... ")

        pred = self.predict(images)

        df_pred = self.get_pred_df(pred)

        df_ans = pd.DataFrame(labels)
        df_ans = df_ans.applymap(lambda x: np.argmax(x))

        df_right = (df_pred == df_ans)

        for num_klass, klass_name in enumerate(df_right):
            logger.debug("Label {} acc: {}/{} = {:.4f}".format(klass_name, np.sum(df_right[klass_name]), len(df_right),
                                                               np.sum(df_right[klass_name]) / len(df_right)))

            # save wrong prediction
            for i in df_pred[klass_name].loc[df_ans[klass_name] != df_pred[klass_name]].index:
                annotation = gen_test.dataset.annotations[i]
                title = f"ans({self.list_class_names[num_klass][df_ans[i]]})_pred({self.list_class_names[num_klass][df_pred[i]]})"

                image = image_processing.draw_annotation(annotation)
                if output_dir is None:
                    image_processing.show(image, title=f"{annotation.filename}\n{title}")
                else:
                    image_processing.save_image(image, os_path.join(output_dir, f"{annotation.filename}_{title}.png"))

        logger.debug(
            "Total acc {}/{} = {:.4f}".format(np.sum(df_right.all(axis=1)), len(df_right),
                                              np.sum(df_right.all(axis=1)) / len(df_right)))


class YoloModel(CNNModel):
    @property
    def num_classes(self):
        return [len(klass) for klass in self.class_names]

    def to_named_classes(self, indexes):
        named_classes = [self.class_names[i].get_pair(indexes[i]) for i in range(len(indexes))]
        if len(named_classes) == 1:
            return named_classes[0]
        return named_classes

    def __init__(self, name, input_size, class_names,
                 training, base_model, anchors, channels=3, grid_factor=32, normalize_input=True, expand_pixel=0):
        super().__init__(name, input_size)
        self.class_names = class_names
        self.training = training
        self.anchors = anchors
        self.anchor_masks = get_anchor_masks(anchors)
        self.channels = channels
        self.img_aug = None
        self.grid_factor = grid_factor
        self.expand_pixel = expand_pixel

        yolo_model = YoloCustomizeModelMultiCategory(input_size[0],
                                                     channels=channels,
                                                     anchors=anchors,
                                                     classes=self.num_classes,
                                                     training=training,
                                                     model=base_model,
                                                     normalize_input=normalize_input)

        self.__dict__.update(yolo_model.__dict__)

    def process_test_data(self, images, convert_BGR=False):
        images = self.img_aug.aug(images)[0]
        images = np.array(images)
        if convert_BGR:
            blue = images[..., 0].copy()
            images[..., 0], images[..., 2] = images[..., 2], blue

        return np.array(images)

    def detect(self, images, convert_BGR=False, to_annotation=False):
        """

        :param images: list of images that has same shape
        :param convert_BGR:
        :param to_annotation:
        :return:
        """
        if not isinstance(images, list) and (isinstance(images, np.ndarray) and images.ndim == 3):
            images = [images]

        reverse_seq = self.get_reverse_seq(images[0].shape[:2])

        processed_images = self.process_test_data(images, convert_BGR)
        preds = self.predict_on_batch(processed_images)
        pred_boxes, scores = decode_nms(preds, self.input_size + (self.channels,))
        pred_labels = convert2_class(scores, self.num_classes)

        # rescale bounding boxes
        def rescale_bbox(pred_box):
            bbox = ia.BoundingBoxesOnImage.from_xyxy_array(pred_box, shape=self.input_size + (self.channels,))
            reversed_bbox = reverse_seq(bounding_boxes=bbox).to_xyxy_array().astype('int32')
            return reversed_bbox

        pred_boxes = [rescale_bbox(pred_box) for pred_box in pred_boxes]
        if not to_annotation:
            return pred_boxes, pred_labels

        annotations = []
        for n_image in range(len(images)):
            pred_klass, pred_score = pred_labels[n_image]
            objs = []

            for n_obj in range(len(pred_klass)):
                # process labels
                named_classes = self.to_named_classes(pred_klass[n_obj])

                # process score
                pred_score_obj = pred_score[n_obj]
                if len(pred_score_obj) == 1:
                    pred_score_obj = pred_score_obj[0]

                # create VOCObject
                objs.append(VOCObject(name=named_classes, bbox=list(pred_boxes[n_image][n_obj]), score=pred_score_obj))

            annotations.append(VOCAnnotation(size=images[n_image].shape[:2], objects=objs))
        return annotations


import numpy as np
import cv2
from cv2 import *
# # enable X11 Forwarding
# # # 1. add DISPLAY to environment variable
# # # 2.
# import matplotlib
# matplotlib.use("Agg")
import matplotlib.pyplot as plt

from utils import os_path


def show(image, convert_BGR=True, title=None):
    if isinstance(image, list):
        show_images(image, convert_BGR, title)
        return 0
    image = image.astype('uint8')
    if len(image.shape) == 2:
        is_bw = True
    else:
        is_bw = False
    if not is_bw:
        if convert_BGR:
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        plt.imshow(image)
    else:
        plt.imshow(image, cmap='gray')
    if title:
        plt.title(title)
    plt.show()


def show_images(images, convert_BGR, titles):
    if titles is not None:
        for image, title in zip(images, titles):
            show(image, convert_BGR, title)
    else:
        for image in images:
            show(image, convert_BGR)


def to_bytes(image, _format=".png"):
    success, encoded_image = cv2.imencode(_format, image)
    image_bytearr = encoded_image.tobytes()
    return image_bytearr


def read_image_from_io(imgData):
    data = np.fromstring(imgData, np.uint8)
    img = cv2.imdecode(data, cv2.IMREAD_COLOR)
    return img


def save_image(image, image_path, make_dir=True, convert_BGR=False):
    if make_dir:
        os_path.make_dir(os_path.os.path.dirname(image_path))

    # if normalized
    if np.max(image) <= 1:
        image = image * 255

    # convert type
    if image.dtype != np.dtype('uint8'):
        image = image.astype('uint8')

    if convert_BGR:
        image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)

    cv2.imwrite(image_path, image)


def draw_box(image, bbox, color=(255, 0, 0), label=None, font_size=1):
    image = np.copy(image)
    H, W = image.shape[:2]
    h_stride, w_stride = H // 60, W // 200
    bbox_label = None
    if isinstance(bbox, list) or isinstance(bbox, tuple) or isinstance(bbox, np.ndarray):
        if len(bbox) == 5:
            x1, y1, x2, y2, bbox_label = bbox
        else:
            x1, y1, x2, y2 = bbox
    elif hasattr(bbox, "name"):  # is object
        x1, y1, x2, y2, bbox_label = int(bbox.x1), int(bbox.y1), int(bbox.x2), int(bbox.y2), bbox.name
    else:
        x1, y1, x2, y2 = int(bbox.x1), int(bbox.y1), int(bbox.x2), int(bbox.y2)

    cv2.rectangle(image, (int(x1), int(y1)), (int(x2), int(y2)), color, 2)
    if label:
        cv2.putText(image, str(label), (x1 + w_stride, y1 - h_stride), cv2.FONT_HERSHEY_COMPLEX_SMALL, font_size, color,
                    1, cv2.LINE_4)
    elif bbox_label:
        cv2.putText(image, str(label), (x1 + w_stride, y1 - h_stride), cv2.FONT_HERSHEY_COMPLEX_SMALL, font_size, color,
                    1, cv2.LINE_4)
    return image


def draw_boxes(image, bboxes, color=(255, 0, 0), labels=None, font_sizes=1):
    if not isinstance(bboxes, list):
        bboxes = [bboxes]

    if not isinstance(labels, list):
        labels = [labels] * len(bboxes)

    if not isinstance(font_sizes, list):
        font_sizes = [font_sizes] * len(bboxes)

    for bbox, label, font_size in zip(bboxes, labels,
                                      font_sizes):  # can be list of lists, list of objects, or list of bboxes
        image = draw_box(image, bbox, color, label=label, font_size=font_size)

    return image


def draw_annotation(annotation, color=(255, 0, 0), image=None):
    if image is None:
        image = imread(annotation.image_path)
    return draw_boxes(image, annotation.objects, color)


def plot_rgb_hist(image, save_path=None):
    # gray scale image
    if image.ndim == 2 or (image.ndim == 3 and image.shape[-1] == 1):
        plt.hist(image.ravel(), 256, [0, 256])
    else:
        color = ('b', 'g', 'r')
        for i, col in enumerate(color):
            histr = cv2.calcHist([image], [i], None, [256], [0, 256])
            plt.plot(histr, color=col)
            plt.xlim([0, 256])
    plt.show()
    if save_path:
        plt.savefig(save_path)


def plot_combined_rgb_hist(image, save_path=None, title=None):
    if isinstance(image, str):
        image = plt.imread(image, format='uint8')
    plt.clf()

    gray = cv2.cvtColor(image.copy(), cv2.COLOR_RGB2GRAY)
    hist_gray = cv2.calcHist([gray], [0], None, [256], [0, 256])

    plt.subplot(221), plt.imshow(gray, 'gray')
    plt.subplot(222), plt.imshow(image)
    plt.subplot(223), plt.plot(hist_gray)
    plt.subplot(224)
    color = ('b', 'g', 'r')
    for i, col in enumerate(color):
        histr = cv2.calcHist([image], [i], None, [256], [0, 256])
        plt.plot(histr, color=col)
        plt.xlim([0, 256])
    plt.xlim([0, 256])

    if title:
        plt.title(title)

    if save_path:
        plt.savefig(save_path)
    else:
        plt.show()






# -*- coding:utf-8 _*-  
"""
@author: Maureen Hsu
@file: flag_utils.py 
@time: 2020/06/02
"""

# python packages
from absl.flags import FLAGS

# 3rd-party packages
from loguru import logger


# self-defined packages

@logger.catch(reraise=True)
def get_flag(attr, default=None):
    try:
        if hasattr(FLAGS, attr):
            return getattr(FLAGS, attr)
    except Exception:
        return default
    return default


@logger.catch(reraise=True)
def log_flag():

    logger.debug(f"Flags: {FLAGS.flags_into_string()}")

